import asyncio
import os
import json
import logging
import argparse
import sys
import crawl
import utils
from blocklists import Blocklists


async def process_urls(
    urls,
    batch_size,
    tracking_domains_list,
    headless=True,
    screenshot=True,
    results_db_file="crawl_results.db",
):
    """
    Start the Playwright browser, run the URLs to test in batches asynchronously
    and write the data to a file.
    """

    return await crawl.crawl_batch(
        urls=urls,
        batch_size=batch_size,
        results_function=crawl.store_crawl_results,
        tracking_domains_list=tracking_domains_list,
        browser_config={"headless": headless, "channel": "msedge"},
        results_db_file=results_db_file,
        screenshot=screenshot,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("url", help="URL or file with URLs to test")
    parser.add_argument(
        "--debug", default=False, action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "--headless",
        default=True,
        type=utils.string_to_boolean,
        const=False,
        nargs="?",
        help="Run browser in headless mode (yes/no)",
    )
    parser.add_argument(
        "--screenshot",
        default=False,
        action="store_true",
        help="Take screenshots of each page before and after consent is given (if consent manager is detected)",
    )
    parser.add_argument(
        "--bootstrap",
        default=False,
        action="store_true",
        help="Force bootstrap (refresh) of blocklists",
    )
    parser.add_argument(
        "--batch_size",
        "-b",
        default=15,
        type=int,
        help="Number of URLs (and browser windows) to run in each batch. Default: 15, increase or decrease depending on your system capacity.",
    )
    parser.add_argument(
        "--show_output",
        "-o",
        default=False,
        action="store_true",
        help="Show output of the last results in terminal (max 25 results)",
    )
    parser.add_argument(
        "--db_file",
        "-db",
        default="crawl_results.db",
        help="Path to crawl results and blocklist database",
    )
    parser.add_argument(
        "--blocklists", "-bf", default=None, help="Path to custom blocklists file"
    )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    if not args.db_file.endswith(".db"):
        args.db_file = args.db_file + ".db"

    if args.blocklists != None:
        if not os.path.isfile(args.blocklists):
            logging.error(f"Blocklists file not found: {args.blocklists}")
            sys.exit(1)

        if not any(
            [args.blocklists.endswith(".yaml"), args.blocklists.endswith(".yml")]
        ):
            logging.error(f"Blocklists file must be a YAML file: {args.blocklists}")
            sys.exit(1)

    if not os.path.isdir("screenshots") and args.screenshot == True:
        os.mkdir("screenshots")

    # List of URLs to test
    if args.url.endswith(".txt"):
        with open(args.url, "r") as f:
            urls = list(
                set(
                    [
                        l.strip().lower()
                        for l in set(f.readlines())
                        if len(l) > 0 and not l.startswith("#")
                    ]
                )
            )

    elif args.url != "":
        urls = args.url.split(",")
    else:
        logging.error("No URL or valid .txt file with URLs to test")

    # Bootstrap blocklists
    blocklists = Blocklists(
        db_file=args.db_file,
        source_file=args.blocklists,
        force_bootstrap=args.bootstrap,
    )

    results = asyncio.run(
        process_urls(
            urls=urls,
            batch_size=args.batch_size,
            tracking_domains_list=blocklists.get_domains(),
            headless=args.headless,
            screenshot=args.screenshot,
            results_db_file=args.db_file,
        )
    )

    if args.show_output and len(results) < 25:
        sys.stdout.write(json.dumps(results, indent=2))

    sys.exit(0)



import os
import json
import logging
import datetime
import re
import base64
import random
import yaml
import asyncio
import sqlite3
from datetime import date, datetime
from pathlib import Path
from playwright.async_api import TimeoutError as PlaywrightTimeoutError
from playwright.async_api import async_playwright
import utils

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
CONSENT_MANAGERS_FILE = f"{MODULE_DIR}/assets/consent_managers.yml"

DEFAULT_UA_STRINGS = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 13_5_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36 Edg/116.0.1938.81"
]


def get_extract_schema():
    return {
        "id": "STRING",
        "url": "STRING",
        "domain_name": "STRING",
        "extraction_datetime": "STRING",
        "cookies_all": "STRING",
        "cookies_no_consent": "STRING",
        "third_party_domains_all": "STRING",
        "third_party_domains_no_consent": "STRING",
        "tracking_domains_all": "STRING",
        "tracking_domains_no_consent": "STRING",
        "consent_manager": "STRING",
        "screenshot_files": "STRING",
        "meta_tags": "STRING",
        "json_ld": "STRING",
        "status": "STRING",
        "status_msg": "STRING",
    }


def get_consent_managers():
    with open(CONSENT_MANAGERS_FILE, "r") as f:
        data = yaml.safe_load(f)
        return data


async def click_consent_manager(page):
    """Retrieve list of potential consent managers and required actions to accept. Then click and return the consent manager."""
    consent_managers = get_consent_managers()

    for cmp in consent_managers:
        parent_locator = page
        locator = None

        for action in cmp["actions"]:
            if action["type"] == "iframe":
                if await parent_locator.locator(action["value"]).count() > 0:
                    parent_locator = parent_locator.frame_locator(action["value"]).first
                else:
                    continue

            elif action["type"] == "css-selector":
                if await parent_locator.locator(action["value"]).first.is_visible():
                    locator = parent_locator.locator(action["value"])
                    break

            elif action["type"] == "css-selector-list":
                for selector in action["value"]:
                    if await parent_locator.locator(selector).first.is_visible():
                        locator = parent_locator.locator(selector)
                        cmp["selector-list-item"] = selector
                        break

            elif action["type"] == "xpath":
                logging.info("XPath not implemented yet.")

        if locator is not None:
            logging.debug(
                f"Found { await locator.count()} elements for consent manager '{cmp['id']}'"
            )

            try:
                # explicit wait for navigation as some pages will reload after accepting cookies
                async with page.expect_navigation(
                    wait_until="networkidle", timeout=15000
                ):
                    await locator.first.click(delay=10)
                    logging.debug(f"Clicked consent manager '{cmp['id']}'")

                    return cmp
            except PlaywrightTimeoutError:
                logging.debug("Timeout, no navigation")
                cmp["status"] = "timeout"
                return cmp

            except Exception as e:
                error_msg = f"Error clicking consent manager '{cmp['id']}': {e}"
                logging.debug(error_msg)
                cmp["status"] = "error"
                cmp["error"] = error_msg
                return cmp

    logging.debug(f"Unable to accept cookies on: {page.url}")
    return {}


async def get_jsonld(page):
    json_ld = []
    for item in await page.locator('script[type="application/ld+json"]').all():
        contents = await item.inner_text()
        try:
            # remove potential CDATA tags
            match = re.search(
                r"//<!\[CDATA\[\s*(.*?)\s*//\]\]>", contents.strip(), re.DOTALL
            )
            if match:
                json_ld.append(json.loads(match.group(1), strict=False))
            else:
                json_ld.append(json.loads(contents.strip(), strict=False))
        except Exception as e:
            logging.debug(f"Unable to parse JSON-LD: {e}")
            json_ld.append({"raw": str(contents), "error": str(e)})

    return json_ld


async def get_meta_tags(page):
    meta_tags = {}
    for tag in await page.locator("meta[name]").all():
        try:
            meta_tags[await tag.get_attribute("name")] = await tag.get_attribute(
                "content"
            )
        except Exception as e:
            logging.debug(f"Unable to get meta tag: {e}")
    return meta_tags


async def crawl_url(
    url,
    browser,
    tracking_domains_list=[],
    screenshot=True,
    device={},
    wait_for_timeout=5000,
):
    """
    Open a new browser context with a URL and extract data about cookies and
    tracking domains before and after consent.
    Returns:
    - All third party domains requested
    - Third party domains requested before consent
    - All tracking domains (based on blocklist)
    - Tracking domains before consent
    - All cookies (name, domain, expiration in days)
    - Cookies set before consent
    - Consent manager that was used on the site
    - Screenshot of the site before consenting
    """
    output = {k: None for k in get_extract_schema().keys()}

    try:
        if not url.startswith("http"):
            url = "http://" + url

        output["url"] = url
        output["extraction_datetime"] = str(datetime.now())

        browser_context = await browser.new_context(
            user_agent=random.choice(DEFAULT_UA_STRINGS),
            viewport={"width": 1366, "height": 768},
            **device,
        )
        await browser_context.add_init_script(
            "Object.defineProperty(navigator, 'webdriver', {get: () => undefined})"
        )

        output["domain_name"] = re.search("(?:https?://)?(?:www.)?([^/]+)", url).group(
            1
        )
        base64_url = base64.urlsafe_b64encode(
            output["domain_name"].encode("ascii")
        ).decode("ascii")

        output["id"] = base64_url

        logging.info(f"Start extracting data from domain {output['domain_name']}")

        req_urls = []

        page = await browser_context.new_page()
        page.on("request", lambda req: req_urls.append(req.url))

        await page.goto(url, wait_until="load", timeout=90000)

        # Do some mouse jiggling to keep some pages happy
        await page.wait_for_timeout(2000)
        await page.mouse.move(543, 123)
        await page.mouse.wheel(0, -123)
        await page.wait_for_timeout(
            wait_for_timeout
        )  # additional wait time just to be sure as consent managers can sometimes take a while to load

        if screenshot:
            await page.screenshot(path=f'./screenshots/screenshot_{output["id"]}.png')
            output["screenshot_files"] = [
                f'./screenshots/screenshot_{output["id"]}.png'
            ]

        logging.debug(f"Retrieving JSON-LD and meta tags on {output['domain_name']}")
        output["json_ld"] = await get_jsonld(page)
        output["meta_tags"] = await get_meta_tags(page)

        # Capture data pre-consent
        thirdparty_requests = list(
            filter(lambda req_url: not output["domain_name"] in req_url, req_urls)
        )
        output["third_party_domains_no_consent"] = list(
            set(
                map(
                    lambda r: re.search("https?://(?:www.)?([^\/]+\.[^\/]+)", r).group(
                        1
                    ),
                    thirdparty_requests,
                )
            )
        )
        output["tracking_domains_no_consent"] = list(
            set(
                [
                    re.search("[^\.]+\.[a-z]+$", d).group(0)
                    for d in output["third_party_domains_no_consent"]
                    if d in tracking_domains_list
                ]
            )
        )

        cookies = await browser_context.cookies()
        output["cookies_no_consent"] = [
            {
                "name": c["name"],
                "domain": c["domain"],
                "expires_days": (
                    date.fromtimestamp(int(c["expires"])) - date.today()
                ).days
                if int(c["expires"]) < 200000000000
                else -1,  # prevent out of range errors for the date
            }
            for c in cookies
        ]

        # try to accept full marketing consent
        logging.debug(
            f"Trying to accept full marketing consent on {output['domain_name']}"
        )
        output["consent_manager"] = await click_consent_manager(page)

        if screenshot and output["consent_manager"].get("status", "") not in [
            "error",
            "",
        ]:
            await page.screenshot(
                path=f'./screenshots/screenshot_{output["id"]}_afterconsent.png'
            )
            output["screenshot_files"].append(
                f'./screenshots/screenshot_{output["id"]}_afterconsent.png'
            )

        thirdparty_requests = list(
            filter(lambda req_url: not output["domain_name"] in req_url, req_urls)
        )
        output["third_party_domains_all"] = list(
            set(
                map(
                    lambda r: re.search("https?://(?:www.)?([^\/]+\.[^\/]+)", r).group(
                        1
                    ),
                    thirdparty_requests,
                )
            )
        )
        output["tracking_domains_all"] = list(
            set(
                [
                    re.search("[^\.]+\.[a-z]+$", d).group(0)
                    for d in output["third_party_domains_all"]
                    if d in tracking_domains_list
                ]
            )
        )

        cookies = await browser_context.cookies()

        output["cookies_all"] = [
            {
                "name": c["name"],
                "domain": c["domain"],
                "expires_days": (
                    date.fromtimestamp(int(c["expires"])) - date.today()
                ).days
                if int(c["expires"]) < 200000000000
                else -1,
            }
            for c in cookies
        ]

        await browser_context.close()

        output["status"] = "success"
        output["status_msg"] = f"Successfully extracted data from {url}"

        return output

    except Exception as e:
        error_msg = f"Error extracting data from {url}: {e}"
        logging.debug(error_msg)

        output["status"] = "error"
        output["status_msg"] = error_msg

        return output


async def crawl_batch(
    urls,
    results_function,
    batch_size=10,
    tracking_domains_list=[],
    browser_config=None,
    screenshot=False,
    **kwargs,
):
    """
    Run the crawler for multiple URLs in batches and apply a (async) function
    to the results. Additional arguments can be passed to the results function.
    """

    if not browser_config:
        browser_config = {"headless": True, "channel": "msedge"}

    async with async_playwright() as p:
        logging.debug("Starting browser")
        browser = await p.chromium.launch(**browser_config)

        for urls_batch in utils.batch(urls, batch_size):
            data = [
                crawl_url(
                    url=url,
                    browser=browser,
                    tracking_domains_list=tracking_domains_list,
                    screenshot=screenshot,
                )
                for url in urls_batch
            ]
            results = [
                r for r in await asyncio.gather(*data)
            ]  # run all urls in parallel
            logging.debug(f"Retrieved batch of {len(data)} URLs")

            await results_function(results, **kwargs)

        await browser.close()

        # return the last batch for convenience
    return results


async def crawl_single(url, tracking_domains_list=[], browser_config=None):
    """Crawl a single URL asynchronously."""

    if not browser_config:
        browser_config = {"headless": True, "channel": "msedge"}

    async with async_playwright() as p:
        logging.debug("Starting browser")
        browser = await p.chromium.launch(**browser_config)

        return await crawl_url(
            url=url, browser=browser, tracking_domains_list=tracking_domains_list
        )


async def store_crawl_results(
    data, table_name="crawl_results", file=None, results_db_file="crawl_results.db"
):
    if file is not None:
        Path.mkdir(Path(results_db_file).parent, exist_ok=True)
        with open(file, "a") as f:
            f.writelines([json.dumps(item) + "\n" for item in data])

    if results_db_file is not None:
        Path.mkdir(Path(results_db_file).parent, exist_ok=True)

        conn = sqlite3.connect(results_db_file)
        c = conn.cursor()
        c.execute(
            f"CREATE TABLE IF NOT EXISTS {table_name} ({','.join([f'{k} TEXT' for k in get_extract_schema().keys()])})"
        )
        conn.commit()

        logging.info(f"Storing {len(data)} records in database")
        c = conn.cursor()
        for d in data:
            logging.info(f"Storing {d['url']}")
            d = {
                k: json.dumps(v) if type(v) in [dict, list, tuple] else v
                for k, v in d.items()
            }
            c.execute(
                f"INSERT INTO {table_name} VALUES ({','.join(['?' for k in d.keys()])})",
                tuple(d.values()),
            )
            conn.commit()

        conn.close()



import requests
import logging
import os
import re
import yaml
from time import time
from pathlib import Path
import sqlite3

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


class Blocklists:
    BLOCKLISTS_FILE = f"{MODULE_DIR}/assets/blocklists.yml"
    DB_FILE = f"{MODULE_DIR}/data/blocklists.db"
    TABLE_NAME = "blocklists"

    def __init__(
        self, db_file=None, source_file=None, max_age_days=7, force_bootstrap=False
    ):
        self.has_blocklists = False
        self.max_age_days = max_age_days
        self.force_bootstrap = force_bootstrap
        self._data = {}
        if db_file:
            self.DB_FILE = db_file
        if source_file:
            self.BLOCKLISTS_FILE = source_file

        self.connection = self.get_connection()

        # create table if not exists
        c = self.connection.cursor()
        c.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.TABLE_NAME} (
                url TEXT,
                ids TEXT,
                last_fetch_time TEXT
            )"""
        )
        self.connection.commit()

        # Get last fetch timestamp and check staleness
        self.last_fetch_timestamp = self.get_last_fetch_timestamp()
        logging.debug(f"Last fetch timestamp: {self.last_fetch_timestamp}")

        if self.last_fetch_timestamp > 0:
            self.has_blocklists = True

        if self.blocklists_older_than(self.max_age_days):
            logging.debug(
                f"Blocklists older than {self.max_age_days} days, fetching new blocklist data"
            )
            self.bootstrap()

        if not self.has_blocklists or self.force_bootstrap:
            logging.debug("Fetching new blocklist data...")
            self.bootstrap()

        self.get_blocklist_data_from_db()

    def bootstrap(self):
        """
        Bootstrap blocklists data from a YAML file.
        """
        self.get_blocklists_file()
        self.get_blocklists_data()
        self.generate_master_list()
        self.last_fetch_timestamp = int(time())
        self.store_blocklists_data()
        self.has_blocklists = True

    def get_domains(self):
        if not self.has_blocklists:
            raise Exception("No blocklists data available.")
        return self._data.keys()

    def get_connection(self):
        """
        Create a connection to a SQLite database.
        """

        db_file = self.DB_FILE

        Path.mkdir(Path(db_file).parent, exist_ok=True)
        logging.debug(f"Connecting to database: {db_file}")
        conn = sqlite3.connect(db_file)

        return conn

    def blocklists_older_than(self, days: int = 7):
        age_in_seconds = int(time()) - self.last_fetch_timestamp
        age_in_days = int(age_in_seconds / (60 * 60 * 24))
        logging.debug(
            f"Bootstrap data age in days: {age_in_days} ( > {days} = {age_in_days > days})"
        )
        return age_in_days > days

    def get_blocklists_file(self):
        """
        Retrieve blocklists and hostfiles from a YAML file.
        """

        with open(self.BLOCKLISTS_FILE, "r") as f:
            self.blocklists = yaml.safe_load(f)

    def get_blocklists_data(self):
        for blocklist in self.blocklists:
            logging.info(
                f"Retrieving blocklist: '{blocklist['name']}' from: {blocklist['url']}"
            )

            if blocklist["type"] == "hostfile":
                data = requests.get(blocklist["url"]).text
                blocklist["data"] = self.parse_hostfile(data)

            elif blocklist["type"] == "blocklist":
                data = requests.get(blocklist["url"]).text.split("\n")
                blocklist["data"] = self.parse_blocklist(data)

            elif blocklist["type"] == "domains":
                blocklist["data"] = requests.get(blocklist["url"]).text.split("\n")

            else:
                raise Exception(f"Unknown blocklist type {blocklist['type']}")

    def parse_hostfile(self, data):
        """
        Retrieve a unique list of domain names from a hostfile with the format similar to:
            '127.0.0.1   domain-name.com'
        """
        domains = []
        for line in data.splitlines():
            if not line.startswith("#"):
                match = re.search(r"^\S+\s+(\S+)$", line)
                if match:
                    domains.append(match.group(1))

        return set(domains)

    def parse_blocklist(self, data):
        """
        Retrieve a unique list of domain names from a blocklist with the format similar to:
            '||domain-name.com^'
        """
        domains = []
        for line in data:
            if line.startswith("||"):
                match = re.search(r"^\|\|([^\/]+)\^$", line)
                if match:
                    domains.append(match.group(1))

        return set(domains)

    def generate_master_list(self):
        """
        Generate a master list of domains from all blocklists.
        """

        for l in self.blocklists:
            for item in l["data"]:
                if item in self._data:
                    self._data[item].append(l["id"])
                else:
                    self._data[item] = [l["id"]]

    def store_blocklists_data(self, table_name="blocklists"):
        """
        Store blocklist data in a SQLite database.
        """
        c = self.connection.cursor()
        c.execute(f"DROP TABLE IF EXISTS {table_name}")
        c.execute(
            f"""
            CREATE TABLE {table_name} (
                url TEXT,
                ids TEXT,
                last_fetch_time TEXT
            )"""
        )
        self.connection.commit()

        c.executemany(
            f"INSERT OR REPLACE INTO {table_name} VALUES (?, ?, ?)",
            (
                [
                    (k, ",".join(self._data[k]), self.last_fetch_timestamp)
                    for k in self._data.keys()
                ]
            ),
        )
        self.connection.commit()

    def get_blocklist_data_from_db(self, table_name="blocklists"):
        """
        Retrieve blocklist data from a SQLite database.
        """
        c = self.connection.cursor()
        c.execute(f"SELECT * FROM {table_name}")
        self._data = dict([(row[0], row[1]) for row in c.fetchall()])

    def get_last_fetch_timestamp(self, table_name="blocklists"):
        """
        Retrieve the last fetch timestamp from a SQLite database.
        """
        c = self.connection.cursor()
        try:
            c.execute(f"SELECT MAX(last_fetch_time) FROM {table_name}")
            return int(c.fetchone()[0])
        except:
            return 0



def batch(iterable, n=1):
    """
    Turn any iterable into a generator of batches of batch size n
    from: https://stackoverflow.com/a/8290508
    """
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def string_to_boolean(v):
    """
    Convert many string options to a boolean value. Useful for argparsing.
    From: https://stackoverflow.com/a/43357954/5761491
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")
